from typing import Optional

from .controllers import CastController
from .controllers import get_app
from .controllers import get_chromecast
from .controllers import get_chromecast_with_ip
from .controllers import get_chromecasts
from .controllers import get_controller
from .error import APIError
from .error import CastError
from .stream_info import StreamInfo


def discover() -> list:
    """Perform discovery of devices present on local network, and return result."""

    return [CattDevice(ip_addr=d.host) for d in get_chromecasts()]


class CattDevice:
    def __init__(self, name: str = "", ip_addr: str = "", lazy: bool = False) -> None:
        """
        Class to easily interface with a ChromeCast.

        :param name: Name of ChromeCast device to interface with.
                     Either name of ip-address must be supplied.
        :param ip_addr: Ip-address of device to interface with.
                       Either name of ip-address must be supplied.
        :param lazy: Postpone first connection attempt to device
                     until first playback action is attempted.
        """

        if not name and not ip_addr:
            raise APIError("Neither name nor ip were supplied")

        self.name = name
        self.ip_addr = ip_addr
        self.uuid = None

        self._cast = None
        # Type comment for compatibility with Py3.5-.
        self._cast_controller = None  # type: Optional[CastController]
        if not lazy:
            self._create_cast()

    def __repr__(self) -> str:
        return "<CattDevice: {}>".format(self.name or self.ip_addr)

    def _create_cast(self) -> None:
        self._cast = get_chromecast_with_ip(self.ip_addr) if self.ip_addr else get_chromecast(self.name)
        if not self._cast:
            raise CastError("Device could not be found")
        self._cast.wait()

        self.name = self._cast.name
        self.ip_addr = self._cast.host
        self.uuid = self._cast.uuid

    def _create_controller(self) -> None:
        self._cast_controller = get_controller(self._cast, get_app("default"))

    @property
    def controller(self):
        if not self._cast:
            self._create_cast()
        if not self._cast_controller:
            self._create_controller()
        return self._cast_controller

    def play_url(self, url: str, resolve: bool = False, block: bool = False) -> None:
        """
        Initiate playback of content.

        :param url: Network location of content.
        :param resolve: Try to resolve location of content stream with Youtube-dl.
                        If this is not set, it is assumed that the url points directly to the stream.
        :param block: Block until playback has stopped,
                      either by end of content being reached, or by interruption.
        """

        if resolve:
            stream = StreamInfo(url)
            url = stream.video_url
        self.controller.prep_app()
        self.controller.play_media_url(url)

        if self.controller.wait_for(["PLAYING"], timeout=10):
            if block:
                self.controller.wait_for(["UNKNOWN", "IDLE"])
        else:
            raise APIError("Playback failed")

    def stop(self) -> None:
        """Stop playback."""

        self.controller.kill()

    def play(self) -> None:
        """Resume playback of paused content."""

        self.controller.prep_control()
        self.controller.play()

    def pause(self) -> None:
        """Pause playback of content."""

        self.controller.prep_control()
        self.controller.pause()

    def seek(self, seconds: int) -> None:
        """
        Seek to arbitrary position in content.

        :param seconds: Position in seconds.
        """

        self.controller.prep_control()
        self.controller.seek(seconds)

    def rewind(self, seconds: int) -> None:
        """
        Seek backwards in content by arbitrary amount of seconds.

        :param seconds: Seek amount in seconds.
        """

        self.controller.prep_control()
        self.controller.rewind(seconds)

    def ffwd(self, seconds: int) -> None:
        """
        Seek forward in content by arbitrary amount of seconds.

        :param seconds: Seek amount in seconds.
        """

        self.controller.prep_control()
        self.controller.ffwd(seconds)

    def volume(self, level: float) -> None:
        """
        Set volume to arbitrary level.

        :param level: Volume level (valid range: 0.0-1.0).
        """
        self.controller.volume(level)

    def volumeup(self, delta: float) -> None:
        """
        Raise volume by arbitrary delta.

        :param delta: Volume delta (valid range: 0.0-1.0).
        """
        self.controller.volumeup(delta)

    def volumedown(self, delta: float) -> None:
        """
        Lower volume by arbitrary delta.

        :param delta: Volume delta (valid range: 0.0-1.0).
        """

        self.controller.volumedown(delta)



class CattError(Exception):
    """ Base exception for catt. """

    pass


class CattUserError(CattError):
    """
    Messages from exceptions that inherit from this class,
    are transformed into error messages to the cli user.
    """

    pass


class StateFileError(CattError):
    """ When a requested state file contains invalid data or nothing. """

    pass


class ListenerError(CattError):
    """ When invalid data is passed to a listener class initializer. """

    pass


class AppSelectionError(CattError):
    """ When invalid data is passed to the app selection mechanism. """

    pass


class PlaylistError(CattError):
    """ When playlist specific operations are attempted with non-playlist info. """

    pass


class APIError(CattError):
    # The scope of this exception is probably going to be a bit broad.
    # We can split it into more exceptions once the api is more mature.
    pass


class SubtitlesError(CattUserError):
    """ When a specified subtitles file cannot be found or its encoding cannot be determined. """

    pass


class CliError(CattUserError):
    """ When the cli user passes invalid commands/options/arguments to catt. """

    pass


class CastError(CattUserError):
    """ When operations are attempted with non-existent or inactive devices. """

    pass


class ControllerError(CattUserError):
    """ When a controller is incapable of the requested action. """

    pass


class ExtractionError(CattUserError):
    """ When the requested media cannot be found or processed by youtube-dl. """

    pass


class FormatError(CattUserError):
    """ When the supplied format filter is invalid or excludes all available formats. """

    pass



# -*- coding: utf-8 -*-
import sys

if sys.version_info.major < 3:
    print("This program requires Python 3 and above to run.")
    sys.exit(1)

__author__ = "Stavros Korokithakis"
__email__ = "hi@stavros.io"
__version__ = "0.11.3"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest

from youtube_dl.utils import DownloadError

from catt.controllers import Cache
from catt.controllers import CCInfo
from catt.stream_info import StreamInfo


def ignore_tmr_failure(func):
    """
    Ignore "Too many requests" failures in a test.

    YouTube will sometimes throttle us and cause the tests to flap. This decorator
    catches the "Too many requests" exceptions in tests and ignores them.
    """

    def wrapper(*args):
        try:
            return func(*args)
        except DownloadError as err:
            if "HTTP Error 429:" in str(err):
                pass
            else:
                raise

    return wrapper


class TestThings(unittest.TestCase):
    @ignore_tmr_failure
    def test_stream_info_youtube_video(self):
        stream = StreamInfo("https://www.youtube.com/watch?v=VZMfhtKa-wo", throw_ytdl_dl_errs=True)
        self.assertIn("https://", stream.video_url)
        self.assertEqual(stream.video_id, "VZMfhtKa-wo")
        self.assertTrue(stream.is_remote_file)
        self.assertEqual(stream.extractor, "youtube")

    @ignore_tmr_failure
    def test_stream_info_youtube_playlist(self):
        stream = StreamInfo(
            "https://www.youtube.com/playlist?list=PL9Z0stL3aRykWNoVQW96JFIkelka_93Sc", throw_ytdl_dl_errs=True
        )
        self.assertIsNone(stream.video_url)
        self.assertEqual(stream.playlist_id, "PL9Z0stL3aRykWNoVQW96JFIkelka_93Sc")
        self.assertTrue(stream.is_playlist)
        self.assertEqual(stream.extractor, "youtube")

    def test_stream_info_other_video(self):
        stream = StreamInfo("https://www.twitch.tv/buddha/clip/OutstandingSquareMallardBudStar")
        self.assertIn("https://", stream.video_url)
        self.assertEqual(stream.video_id, "471296669")
        self.assertTrue(stream.is_remote_file)
        self.assertEqual(stream.extractor, "twitch")

    def test_cache(self):
        cache = Cache()
        cache.set_data("name", CCInfo("192.168.0.6", 8009, "Fake Factory Inc.", "Fakecast", "fake"))
        self.assertEqual(
            cache.get_data("name").all_info,
            {
                "ip": "192.168.0.6",
                "port": 8009,
                "manufacturer": "Fake Factory Inc.",
                "model_name": "Fakecast",
                "cast_type": "fake",
            },
        )

        cache.clear()
        cache = Cache()
        self.assertEqual(cache.get_data("name"), None)
        cache.clear()


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys

from catt import __version__

if sys.version_info.major < 3:
    print("This program requires Python 3 and above to run.")
    sys.exit(1)


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open("README.rst") as readme_file:
    readme = readme_file.read()

requirements = [
    "youtube-dl>=2020.6.6",
    "PyChromecast~=6.0",
    "Click>=7.1.2",
    "ifaddr>=0.1.7",
    "requests>=2.23.0",
    'typing;python_version<"3.5"',
]

test_requirements = []  # type: ignore

setup(
    name="catt",
    version=__version__,
    description="Cast All The Things allows you to send videos from many, many online sources to your Chromecast.",
    long_description=readme,
    author="Stavros Korokithakis",
    author_email="hi@stavros.io",
    url="https://github.com/skorokithakis/catt",
    packages=["catt"],
    package_dir={"catt": "catt"},
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords="chromecast cast catt cast_all_the_things",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    test_suite="tests",
    tests_require=test_requirements,
    python_requires=">=3.4",
    entry_points={"console_scripts": ["catt=catt.cli:main"]},
)



import io
import re
import socketserver
import sys
import time
import traceback
from http.server import BaseHTTPRequestHandler
from pathlib import Path
from typing import Optional
from typing import Tuple

BYTE_RANGE_RE = re.compile(r"bytes=(\d+)-(\d+)?$")


def copy_byte_range(
    infile: io.BufferedIOBase,
    outfile: io.BufferedIOBase,
    start: int = None,
    stop: int = None,
    bufsize: int = 16 * 1024,
):
    """Like shutil.copyfileobj, but only copy a range of the streams.

    Both start and stop are inclusive.
    """
    if start is not None:
        infile.seek(start)
    while True:
        to_read = min(bufsize, stop + 1 - infile.tell() if stop else bufsize)
        buf = infile.read(to_read)
        if not buf:
            break
        outfile.write(buf)


def parse_byte_range(byte_range: str) -> Tuple[Optional[int], Optional[int]]:
    """Returns the two numbers in 'bytes=123-456' or throws ValueError.

    The last number or both numbers may be None.
    """
    if byte_range.strip() == "":
        return None, None

    match = BYTE_RANGE_RE.match(byte_range)
    if not match:
        raise ValueError("Invalid byte range {}".format(byte_range))

    first, last = [int(x) if x else None for x in match.groups()]
    assert first is not None
    if last and last < first:
        raise ValueError("Invalid byte range {}".format(byte_range))
    return first, last


def serve_file(
    filename: str, address: str = "", port: int = 45114, content_type=None, single_req=False,
):
    class FileHandler(BaseHTTPRequestHandler):
        def format_size(self, size):
            for size_unity in ["B", "KB", "MB", "GB", "TB"]:
                if size < 1024:
                    return size, size_unity
                size = size / 1024
            return size * 1024, size_unity

        def log_message(self, format, *args, **kwargs):
            size, size_unity = self.format_size(stats.st_size)
            format += " {} - {:0.2f} {}".format(content_type, size, size_unity)
            return super(FileHandler, self).log_message(format, *args, **kwargs)

        def do_GET(self):  # noqa
            if "Range" not in self.headers:
                first, last = 0, stats.st_size
            else:
                try:
                    first, last = parse_byte_range(self.headers["Range"])
                except ValueError:
                    self.send_error(400, "Invalid byte range")
                    return None

            if last is None or last >= stats.st_size:
                last = stats.st_size - 1
            response_length = last - first + 1

            try:
                if "Range" not in self.headers:
                    self.send_response(200)
                else:
                    self.send_response(206)
                    self.send_header(
                        "Content-Range", "bytes {}-{}/{}".format(first, last, stats.st_size),
                    )

                self.send_header("Accept-Ranges", "bytes")
                self.send_header("Content-type", content_type)
                self.send_header("Content-Length", str(response_length))
                self.send_header("Access-Control-Allow-Origin", "*")
                self.send_header(
                    "Last-Modified", time.strftime("%a %d %b %Y %H:%M:%S GMT", time.localtime(stats.st_mtime)),
                )
                self.end_headers()

                mediafile = open(str(mediapath), "rb")
                copy_byte_range(mediafile, self.wfile, first, last)
            except ConnectionResetError:
                # This is supposed to happen when the Chromecast seeks or stops.
                pass
            except BrokenPipeError:
                print(
                    "Device disconnected while playing. Please check that the video file is compatible with the device.",
                    file=sys.stderr,
                )
            except:  # noqa
                traceback.print_exc()

            mediafile.close()

    if content_type is None:
        content_type = "video/mp4"

    mediapath = Path(filename)
    stats = mediapath.stat()

    httpd = socketserver.TCPServer((address, port), FileHandler)
    if single_req:
        httpd.handle_request()
    else:
        httpd.serve_forever()
    httpd.server_close()
